// ****************************************************************************
// 
// FLV Extract
// Copyright (C) 2006-2008  Moitah (moitah@yahoo.com)
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// ****************************************************************************

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace JDP {
	interface IAudioWriter {
		void WriteChunk(byte[] chunk);
		void Finish();
		string Path { get; }
	}

	interface IVideoWriter {
		void WriteChunk(byte[] chunk, uint timeStamp, int frameType);
		void Finish(FractionUInt32 averageFrameRate);
		string Path { get; }
	}

	class FLVFile : IDisposable {
		string _path, _pathBase;
		FileStream _fs;
		long _fileOffset, _fileLength;
		IAudioWriter _audioWriter;
		IVideoWriter _videoWriter;
		TimeCodeWriter _timeCodeWriter;
		List<uint> _videoTimeStamps;
		bool _extractAudio, _extractVideo, _extractTimeCodes;
		bool _extractedAudio, _extractedVideo, _extractedTimeCodes;
		FractionUInt32? _averageFrameRate, _trueFrameRate;
		List<string> _warnings;

		public FLVFile(string path) {
			_path = path;
			_warnings = new List<string>();
			_fs = new FileStream(path, FileMode.Open, FileAccess.Read, FileShare.Read, 65536);
			_fileOffset = 0;
			_fileLength = _fs.Length;
		}

		public void Close() {
			Dispose();
		}

		public void Dispose() {
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing) {
			if (disposing) {
				if (_fs != null) {
					_fs.Close();
					_fs = null;
				}
				CloseOutput(null, true);
			}
		}

		public FractionUInt32? AverageFrameRate {
			get { return _averageFrameRate; }
		}

		public FractionUInt32? TrueFrameRate {
			get { return _trueFrameRate; }
		}

		public string[] Warnings {
			get { return _warnings.ToArray(); }
		}

		public bool ExtractedAudio {
			get { return _extractedAudio; }
		}

		public bool ExtractedVideo {
			get { return _extractedVideo; }
		}

		public bool ExtractedTimeCodes {
			get { return _extractedTimeCodes; }
		}

		public void ExtractStreams(bool extractAudio, bool extractVideo, bool extractTimeCodes) {
			uint dataOffset, flags, prevTagSize;
			string ext;

			ext = Path.GetExtension(_path).ToLower();
			if ((ext == ".avi") || (ext == ".mp3") || (ext == ".txt")) {
				// Can't have the same extension as the files we output
				throw new Exception("Please change the extension of your FLV file.");
			}
			_pathBase = _path.Substring(0, _path.Length - ext.Length);

			_extractAudio = extractAudio;
			_extractVideo = extractVideo;
			_extractTimeCodes = extractTimeCodes;
			_videoTimeStamps = new List<uint>();

			Seek(0);
			if (ReadUInt32() != 0x464C5601) {
				throw new Exception("This isn't a FLV file.");
			}

			flags = ReadUInt8();
			dataOffset = ReadUInt32();

			Seek(dataOffset);

			prevTagSize = ReadUInt32();
			while (_fileOffset < _fileLength) {
				if (!ReadTag()) break;
				if ((_fileLength - _fileOffset) < 4) break;
				prevTagSize = ReadUInt32();
			}

			_averageFrameRate = CalculateAverageFrameRate();
			_trueFrameRate = CalculateTrueFrameRate();

			CloseOutput(_averageFrameRate, false);
		}

		private void CloseOutput(FractionUInt32? averageFrameRate, bool disposing) {
			if (_videoWriter != null) {
				_videoWriter.Finish(averageFrameRate ?? new FractionUInt32(25, 1));
				if (disposing && (_videoWriter.Path != null)) {
					try { File.Delete(_videoWriter.Path); }
					catch { }
				}
				_videoWriter = null;
			}
			if (_audioWriter != null) {
				_audioWriter.Finish();
				if (disposing && (_audioWriter.Path != null)) {
					try { File.Delete(_audioWriter.Path); }
					catch { }
				}
				_audioWriter = null;
			}
			if (_timeCodeWriter != null) {
				_timeCodeWriter.Finish();
				if (disposing && (_timeCodeWriter.Path != null)) {
					try { File.Delete(_timeCodeWriter.Path); }
					catch { }
				}
				_timeCodeWriter = null;
			}
		}

		private bool ReadTag() {
			uint tagType, dataSize, timeStamp, streamID, mediaInfo;
			byte[] data;

			if ((_fileLength - _fileOffset) < 11) {
				return false;
			}

			// Read tag header
			tagType = ReadUInt8();
			dataSize = ReadUInt24();
			timeStamp = ReadUInt24();
			timeStamp |= ReadUInt8() << 24; 
			streamID = ReadUInt24();

			// Read tag data
			if (dataSize == 0) {
				return true;
			}
			if ((_fileLength - _fileOffset) < dataSize) {
				return false;
			}
			mediaInfo = ReadUInt8();
			dataSize -= 1;
			data = ReadBytes((int)dataSize);

			if (tagType == 0x8) {  // Audio
				if (_audioWriter == null) {
					_audioWriter = _extractAudio ? GetAudioWriter(mediaInfo) : new DummyAudioWriter();
					_extractedAudio = !(_audioWriter is DummyAudioWriter);
				}
				_audioWriter.WriteChunk(data);
			}
			else if (tagType == 0x9) { // Video
				if (_videoWriter == null) {
					_videoWriter = _extractVideo ? GetVideoWriter(mediaInfo) : new DummyVideoWriter();
					_extractedVideo = !(_videoWriter is DummyVideoWriter);
				}
				if (_timeCodeWriter == null) {
					_timeCodeWriter = new TimeCodeWriter(_extractTimeCodes ? (_pathBase + ".txt") : null);
					_extractedTimeCodes = _extractTimeCodes;
				}
				_videoTimeStamps.Add(timeStamp);
				_videoWriter.WriteChunk(data, timeStamp, (int)((mediaInfo & 0xF0) >> 4));
				_timeCodeWriter.Write(timeStamp);
			}

			return true;
		}

		private IAudioWriter GetAudioWriter(uint mediaInfo) {
			uint format = mediaInfo >> 4;
			uint rate = (mediaInfo >> 2) & 0x3;
			uint bits = (mediaInfo >> 1) & 0x1;
			uint chans = mediaInfo & 0x1;

			if (format == 2) { // MP3
				return new MP3Writer(_pathBase + ".mp3");
			}
			else if (format == 0) { // Uncompressed
				int sampleRate = 0;
				switch (rate) {
					case 0: sampleRate =  5512; break;
					case 1: sampleRate = 11025; break;
					case 2: sampleRate = 22050; break;
					case 3: sampleRate = 44100; break;
				}
				return new WAVWriter(_pathBase + ".wav", (bits == 1) ? 16 : 8,
					(chans == 1) ? 2 : 1, sampleRate);
			}
			else {
				string typeStr;

				if (format == 1)
					typeStr = "ADPCM";
				else if ((format == 5) || (format == 6))
					typeStr = "Nellymoser";
				else
					typeStr = "format=" + format.ToString();

				_warnings.Add("Unable to extract audio (" + typeStr + " is unsupported).");

				return new DummyAudioWriter();
			}
		}

		private IVideoWriter GetVideoWriter(uint mediaInfo) {
			uint codecID = mediaInfo & 0x0F;

			if ((codecID == 2) || (codecID == 4) || (codecID == 5)) {
				return new AVIWriter(_pathBase + ".avi", (int)codecID);
			}
			else {
				string typeStr;

				if (codecID == 3)
					typeStr = "Screen";
				else if (codecID == 6)
					typeStr = "Screen2";
				else
					typeStr = "codecID=" + codecID.ToString();

				_warnings.Add("Unable to extract video (" + typeStr + " is unsupported).");

				return new DummyVideoWriter();
			}
		}

		private FractionUInt32? CalculateAverageFrameRate() {
			FractionUInt32 frameRate;
			int frameCount = _videoTimeStamps.Count;

			if (frameCount > 1) {
				frameRate.N = (uint)(frameCount - 1) * 1000;
				frameRate.D = _videoTimeStamps[frameCount - 1] - _videoTimeStamps[0];
				frameRate.Reduce();
				return frameRate;
			}
			else {
				return null;
			}
		}

		private FractionUInt32? CalculateTrueFrameRate() {
			FractionUInt32 frameRate;
			Dictionary<uint, uint> deltaCount = new Dictionary<uint, uint>();
			int i, threshold;
			uint delta, count, minDelta;

			// Calculate the distance between the timestamps, count how many times each delta appears
			for (i = 1; i < _videoTimeStamps.Count; i++) {
				int deltaS = (int)((long)_videoTimeStamps[i] - (long)_videoTimeStamps[i - 1]);

				if (deltaS < 0) continue;
				delta = (uint)deltaS;

				if (deltaCount.ContainsKey(delta)) {
					deltaCount[delta] += 1;
				}
				else {
					deltaCount.Add(delta, 1);
				}
			}

			threshold = _videoTimeStamps.Count / 10;
			minDelta = UInt32.MaxValue;

			// Find the smallest delta that made up at least 10% of the frames (grouping in delta+1
			// because of rounding, e.g. a NTSC video will have deltas of 33 and 34 ms)
			foreach (KeyValuePair<uint, uint> deltaItem in deltaCount) {
				delta = deltaItem.Key;
				count = deltaItem.Value;

				if (deltaCount.ContainsKey(delta + 1)) {
					count += deltaCount[delta + 1];
				}

				if ((count >= threshold) && (delta < minDelta)) {
					minDelta = delta;
				}
			}

			// Calculate the frame rate based on the smallest delta, and delta+1 if present
			if (minDelta != UInt32.MaxValue) {
				uint totalTime, totalFrames;

				count = deltaCount[minDelta];
				totalTime = minDelta * count;
				totalFrames = count;

				if (deltaCount.ContainsKey(minDelta + 1)) {
					count = deltaCount[minDelta + 1];
					totalTime += (minDelta + 1) * count;
					totalFrames += count;
				}

				if (totalTime != 0) {
					frameRate.N = totalFrames * 1000;
					frameRate.D = totalTime;
					frameRate.Reduce();
					return frameRate;
				}
			}

			// Unable to calculate frame rate
			return null;
		}

		private void Seek(long offset) {
			_fs.Seek(offset, SeekOrigin.Begin);
			_fileOffset = offset;
		}

		private uint ReadUInt8() {
			_fileOffset += 1;
			return (uint)_fs.ReadByte();
		}

		private uint ReadUInt24() {
			byte[] x = new byte[4];
			_fs.Read(x, 1, 3);
			_fileOffset += 3;
			return ByteOrder.Reverse(BitConverter.ToUInt32(x, 0));
		}

		private uint ReadUInt32() {
			byte[] x = new byte[4];
			_fs.Read(x, 0, 4);
			_fileOffset += 4;
			return ByteOrder.Reverse(BitConverter.ToUInt32(x, 0));
		}

		private byte[] ReadBytes(int length) {
			byte[] buff = new byte[length];
			_fs.Read(buff, 0, length);
			_fileOffset += length;
			return buff;
		}
	}

	class DummyAudioWriter : IAudioWriter {
		public DummyAudioWriter() {
		}

		public void WriteChunk(byte[] chunk) {
		}

		public void Finish() {
		}

		public string Path {
			get {
				return null;
			}
		}
	}

	class DummyVideoWriter : IVideoWriter {
		public DummyVideoWriter() {
		}

		public void WriteChunk(byte[] chunk, uint timeStamp, int frameType) {
		}

		public void Finish(FractionUInt32 averageFrameRate) {
		}

		public string Path {
			get {
				return null;
			}
		}
	}

	class MP3Writer : IAudioWriter {
		string _path;
		FileStream _fs;

		public MP3Writer(string path) {
			_path = path;
			_fs = new FileStream(path, FileMode.Create, FileAccess.Write, FileShare.Read, 65536);
		}

		public void WriteChunk(byte[] chunk) {
			_fs.Write(chunk, 0, chunk.Length);
		}

		public void Finish() {
			_fs.Close();
		}

		public string Path {
			get {
				return _path;
			}
		}
	}

	class WAVWriter : IAudioWriter {
		string _path;
		WAVTools.WAVWriter _wr;
		int blockAlign;

		public WAVWriter(string path, int bitsPerSample, int channelCount, int sampleRate) {
			_path = path;
			_wr = new WAVTools.WAVWriter(path, bitsPerSample, channelCount, sampleRate);
			blockAlign = (bitsPerSample / 8) * channelCount;
		}

		public void WriteChunk(byte[] chunk) {
			_wr.Write(chunk, chunk.Length / blockAlign);
		}

		public void Finish() {
			_wr.Close();
		}

		public string Path {
			get {
				return _path;
			}
		}
	}

	class AVIWriter : IVideoWriter {
		string _path;
		BinaryWriter _bw;
		int _codecID;
		int _width, _height, _frameCount;
		uint _moviDataSize, _indexChunkSize;
		List<uint> _index;

		// Chunk:          Off:  Len:
		//
		// RIFF AVI          0    12
		//   LIST hdrl      12    12
		//     avih         24    64
		//     LIST strl    88    12
		//       strh      100    64
		//       strf      164    48
		//   LIST movi     212    12
		//     (frames)    224   ???
		//   idx1          ???   ???

		public AVIWriter(string path, int codecID) {
			if ((codecID != 2) && (codecID != 4) && (codecID != 5)) {
				throw new Exception("Unsupported video codec.");
			}
			_codecID = codecID;

			_path = path;

			FileStream fs = new FileStream(path, FileMode.Create, FileAccess.Write, FileShare.Read);

			_bw = new BinaryWriter(fs);

			WriteFourCC("RIFF");
			_bw.Write((uint)0); // chunk size
			WriteFourCC("AVI ");

			WriteFourCC("LIST");
			_bw.Write((uint)192);
			WriteFourCC("hdrl");

			WriteFourCC("avih");
			_bw.Write((uint)56);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0x10);
			_bw.Write((uint)0); // frame count
			_bw.Write((uint)0);
			_bw.Write((uint)1);
			_bw.Write((uint)0);
			_bw.Write((uint)0); // width
			_bw.Write((uint)0); // height
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);

			WriteFourCC("LIST");
			_bw.Write((uint)116);
			WriteFourCC("strl");

			WriteFourCC("strh");
			_bw.Write((uint)56);
			WriteFourCC("vids");
			WriteFourCC(CodecFourCC);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0); // frame rate denominator
			_bw.Write((uint)0); // frame rate numerator
			_bw.Write((uint)0);
			_bw.Write((uint)0); // frame count
			_bw.Write((uint)0);
			_bw.Write((int)-1);
			_bw.Write((uint)0);
			_bw.Write((ushort)0);
			_bw.Write((ushort)0);
			_bw.Write((ushort)0); // width
			_bw.Write((ushort)0); // height

			WriteFourCC("strf");
			_bw.Write((uint)40);
			_bw.Write((uint)40);
			_bw.Write((uint)0); // width
			_bw.Write((uint)0); // height
			_bw.Write((ushort)1);
			_bw.Write((ushort)24);
			WriteFourCC(CodecFourCC);
			_bw.Write((uint)0); // biSizeImage
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);
			_bw.Write((uint)0);

			WriteFourCC("LIST");
			_bw.Write((uint)0); // chunk size
			WriteFourCC("movi");

			_index = new List<uint>();
		}

		public void WriteChunk(byte[] chunk, uint timeStamp, int frameType) {
			int offset, len;

			offset = 0;
			if (_codecID == 4) offset = 1;
			if (_codecID == 5) offset = 4;
			len = Math.Max(chunk.Length - offset, 0);

			_index.Add((frameType == 1) ? (uint)0x10 : (uint)0);
			_index.Add(_moviDataSize + 4);
			_index.Add((uint)len);

			if ((_width == 0) && (_height == 0)) {
				GetFrameSize(chunk);
			}

			WriteFourCC("00dc");
			_bw.Write(len);
			_bw.Write(chunk, offset, len);

			if ((len % 2) != 0) {
				_bw.Write((byte)0);
				len++;
			}
			_moviDataSize += (uint)len + 8;
			_frameCount++;
		}

		private void GetFrameSize(byte[] chunk) {
			if (_codecID == 2) {
				// Reference: flv_h263_decode_picture_header from libavcodec's h263.c

				if (chunk.Length < 10) return;

				if ((chunk[0] != 0) || (chunk[1] != 0)) {
					return;
				}

				ulong x = ByteOrder.Reverse(BitConverter.ToUInt64(chunk, 2));
				int format;

				if (ReadBits(ref x, 1) != 1) {
					return;
				}
				ReadBits(ref x, 5);
				ReadBits(ref x, 8);

				format = ReadBits(ref x, 3);
				switch (format) {
					case 0:
						_width = ReadBits(ref x, 8);
						_height = ReadBits(ref x, 8);
						break;
					case 1:
						_width = ReadBits(ref x, 16);
						_height = ReadBits(ref x, 16);
						break;
					case 2:
						_width = 352;
						_height = 288;
						break;
					case 3:
						_width = 176;
						_height = 144;
						break;
					case 4:
						_width = 128;
						_height = 96;
						break;
					case 5:
						_width = 320;
						_height = 240;
						break;
					case 6:
						_width = 160;
						_height = 120;
						break;
					default:
						return;
				}
			}
			else if ((_codecID == 4) || (_codecID == 5)) {
				// Reference: vp6_parse_header from libavcodec's vp6.c

				int skip = (_codecID == 4) ? 1 : 4;
				if (chunk.Length < (skip + 8)) return;
				ulong x = ByteOrder.Reverse(BitConverter.ToUInt64(chunk, skip));

				int deltaFrameFlag = ReadBits(ref x, 1);
				int quant = ReadBits(ref x, 6);
				int separatedCoeffFlag = ReadBits(ref x, 1);
				int subVersion = ReadBits(ref x, 5);
				int filterHeader = ReadBits(ref x, 2);
				int interlacedFlag = ReadBits(ref x, 1);

				if (deltaFrameFlag != 0) {
					return;
				}
				if ((separatedCoeffFlag != 0) || (filterHeader == 0)) {
					ReadBits(ref x, 16);
				}

				_height = ReadBits(ref x, 8) * 16;
				_width = ReadBits(ref x, 8) * 16;
			}
		}

		private string CodecFourCC {
			get {
				if (_codecID == 2) {
					return "FLV1";
				}
				if ((_codecID == 4) || (_codecID == 5)) {
					return "VP6F";
				}
				return "NULL";
			}
		}

		private int ReadBits(ref ulong x, int length) {
			int r = (int)(x >> (64 - length));
			x <<= length;
			return r;
		}

		private void WriteIndexChunk() {
			uint indexDataSize = (uint)_frameCount * 16;

			WriteFourCC("idx1");
			_bw.Write(indexDataSize);

			for (int i = 0; i < _frameCount; i++) {
				WriteFourCC("00dc");
				_bw.Write(_index[(i * 3) + 0]);
				_bw.Write(_index[(i * 3) + 1]);
				_bw.Write(_index[(i * 3) + 2]);
			}

			_indexChunkSize = indexDataSize + 8;
		}

		public void Finish(FractionUInt32 averageFrameRate) {
			WriteIndexChunk();

			_bw.BaseStream.Seek(4, SeekOrigin.Begin);
			_bw.Write((uint)(224 + _moviDataSize + _indexChunkSize - 8));

			_bw.BaseStream.Seek(24 + 8, SeekOrigin.Begin);
			_bw.Write((uint)0);
			_bw.BaseStream.Seek(12, SeekOrigin.Current);
			_bw.Write((uint)_frameCount);
			_bw.BaseStream.Seek(12, SeekOrigin.Current);
			_bw.Write((uint)_width);
			_bw.Write((uint)_height);

			_bw.BaseStream.Seek(100 + 28, SeekOrigin.Begin);
			_bw.Write((uint)averageFrameRate.D);
			_bw.Write((uint)averageFrameRate.N);
			_bw.BaseStream.Seek(4, SeekOrigin.Current);
			_bw.Write((uint)_frameCount);
			_bw.BaseStream.Seek(16, SeekOrigin.Current);
			_bw.Write((ushort)_width);
			_bw.Write((ushort)_height);

			_bw.BaseStream.Seek(164 + 12, SeekOrigin.Begin);
			_bw.Write((uint)_width);
			_bw.Write((uint)_height);
			_bw.BaseStream.Seek(8, SeekOrigin.Current);
			_bw.Write((uint)(_width * _height * 6));

			_bw.BaseStream.Seek(212 + 4, SeekOrigin.Begin);
			_bw.Write((uint)(_moviDataSize + 4));

			_bw.Close();
		}

		private void WriteFourCC(string fourCC) {
			byte[] bytes = System.Text.Encoding.ASCII.GetBytes(fourCC);
			if (bytes.Length != 4) {
				throw new Exception("Invalid FourCC length.");
			}
			_bw.Write(bytes);
		}

		public string Path {
			get {
				return _path;
			}
		}
	}

	class TimeCodeWriter {
		string _path;
		StreamWriter _sw;

		public TimeCodeWriter(string path) {
			_path = path;
			if (path != null) {
				_sw = new StreamWriter(path, false, Encoding.ASCII);
				_sw.WriteLine("# timecode format v2");
			}
		}

		public void Write(uint timeStamp) {
			if (_sw != null) {
				_sw.WriteLine(timeStamp);
			}
		}

		public void Finish() {
			if (_sw != null) {
				_sw.Close();
				_sw = null;
			}
		}

		public string Path {
			get {
				return _path;
			}
		}
	}

	struct FractionUInt32 {
		public uint N;
		public uint D;

		public FractionUInt32(uint n, uint d) {
			N = n;
			D = d;
		}

		public double ToDouble() {
			return (double)N / (double)D;
		}

		public void Reduce() {
			uint gcd = GCD(N, D);
			N /= gcd;
			D /= gcd;
		}

		private uint GCD(uint a, uint b) {
			uint r;

			while (b != 0) {
				r = a % b;
				a = b;
				b = r;
			}

			return a;
		}

		public override string ToString() {
			return ToString(true);
		}

		public string ToString(bool full) {
			if (full) {
				return ToDouble().ToString() + " (" + N.ToString() + "/" + D.ToString() + ")";
			}
			else {
				return ToDouble().ToString("0.####");
			}
		}
	}
}